#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Simplified Flash Unit Model, only for IAPWS with mixed state.
Phase separator: inlet water/steam mixture is separated into liquid and vapor
streams.

Created: August 21 2020

Created on Thu Aug 18 12:59:50 2020 by Boiler Team (J. Ma, M. Zamarripa)
"""
import pytest

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom

# Import Unit Model Modules
from idaes.models.properties import iapws95

from idaes.models_extra.power_generation.unit_models.helm.phase_separator import (
    HelmPhaseSeparator,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.fixture
def build_phase_separator():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = HelmPhaseSeparator(property_package=m.fs.properties)
    return m


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_basic_build(build_phase_separator):
    """Make a model and make sure it doesn't throw exception"""
    m = build_phase_separator
    assert degrees_of_freedom(m) == 3
    # Check unit config arguments
    assert len(m.fs.unit.config) == 4
    assert m.fs.unit.config.property_package is m.fs.properties


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize(build_phase_separator):
    state_args_water_steam = {
        "flow_mol": 1.5e5,  # mol/s
        "pressure": 1.2e7,  # Pa
        "enth_mol": 28365.2608,
    }  # j/mol
    initialization_tester(
        build_phase_separator, dof=3, state_args_water_steam=state_args_water_steam
    )


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_wflash(build_phase_separator):
    m = build_phase_separator
    state_args_water_steam = {
        "flow_mol": 1.5e5,  # mol/s
        "pressure": 1.2e7,  # Pa
        "enth_mol": 28365.2608,
    }  # j/mol
    m.fs.unit.initialize(state_args_water_steam=state_args_water_steam)
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    solver.solve(m, tee=True)
    Fin = pyo.value(m.fs.unit.mixed_state[0].flow_mol)
    FL = pyo.value(m.fs.unit.liq_state[0].flow_mol)
    FV = pyo.value(m.fs.unit.vap_state[0].flow_mol)
    assert degrees_of_freedom(m) == 0
    assert pytest.approx((FL + FV), abs=1e-3) == Fin

    assert pytest.approx(
        pyo.value(
            m.fs.unit.mixed_state[0].flow_mol * m.fs.unit.mixed_state[0].vapor_frac
        ),
        abs=1e-3,
    ) == pyo.value(m.fs.unit.vap_state[0].flow_mol)

    assert pytest.approx(
        pyo.value(m.fs.unit.mixed_state[0].enth_mol_phase["Vap"]), abs=1e-3
    ) == pyo.value(m.fs.unit.vap_state[0].enth_mol)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_get_stream_table_contents(build_phase_separator):
    stable = build_phase_separator.fs.unit._get_stream_table_contents()

    expected = {
        "Units": {
            "Mass Flow": getattr(pyo.units.pint_registry, "kg/s"),
            "Molar Flow": getattr(pyo.units.pint_registry, "mol/s"),
            "Molar Enthalpy": getattr(pyo.units.pint_registry, "J/mol"),
            "P": getattr(pyo.units.pint_registry, "Pa"),
            "T": getattr(pyo.units.pint_registry, "K"),
            "Vapor Fraction": getattr(pyo.units.pint_registry, "dimensionless"),
        },
        "Inlet": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
        "Vapor Outlet": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
        "Liquid Outlet": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
    }

    assert stable.to_dict() == expected



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
A simple pipeline unit model. We include a bulk momentum balance and leverage
the 1D control volume for a phase material balance. For now, we assume the
pipeline is isothermal.

Data sources:
    [1] Stochastic Optimal Control Model for Natural Gas Network
        Operations. V. Zavala, 2014, Comp. Chem. Eng.

"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from pyomo.common.config import ConfigValue
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.var import Var
from pyomo.core.base.param import Param
from pyomo.core.base.units_container import units as pyunits
from pyomo.core.base.reference import Reference
from pyomo.dae.diffvar import DerivativeVar
from pyomo.core.expr import log10
from pyomo.core.expr.numvalue import value as pyo_value

from idaes.core import (
    declare_process_block_class,
    ControlVolume1DBlock,
    StateBlock,
    UnitModelBlockData,
)
from idaes.core.util.config import (
    is_physical_parameter_block,
    is_transformation_method,
    is_transformation_scheme,
)
from idaes.core.util.constants import Constants


EXPLICIT_DISCRETIZATION_SCHEMES = {
    "FORWARD",
}
IMPLICIT_DISCRETIZATION_SCHEMES = {
    "BACKWARD",
    "LAGRANGE-RADAU",
}
UNSUPPORTED_DISCRETIZATION_SCHEMES = {
    "CENTRAL",
    "LAGRANGE-LEGENDRE",
}


@declare_process_block_class("GasPipeline")
class GasPipelineData(UnitModelBlockData):

    CONFIG = UnitModelBlockData.CONFIG()

    CONFIG.declare(
        "property_package",
        ConfigValue(default=None, domain=is_physical_parameter_block),
    )
    CONFIG.declare(
        "transformation_method",
        ConfigValue(
            default="dae.finite_difference",
            domain=is_transformation_method,
            description="DAE transformation method",
            doc="""Method to use to transform domain. Must be a method recognised
by the Pyomo TransformationFactory.""",
        ),
    )
    CONFIG.declare(
        "transformation_scheme",
        ConfigValue(
            default="FORWARD",
            domain=is_transformation_scheme,
            description="DAE transformation scheme",
            doc="""Scheme to use when transforming domain. See Pyomo
documentation for supported schemes.""",
        ),
    )
    CONFIG.declare(
        "finite_elements",
        ConfigValue(
            default=1,
            domain=int,
            description="Number of finite elements",
            doc="""Number of finite elements to use in transformation (equivalent
to Pyomo nfe argument).""",
        ),
    )
    CONFIG.declare(
        "collocation_points",
        ConfigValue(
            default=None,
            domain=int,
            description="Number of collocation points",
            doc="""Number of collocation points to use (equivalent to Pyomo ncp
argument).""",
        ),
    )

    def build(self):
        super(GasPipelineData, self).build()

        # self.config is the ConfigBlock "instantiated" from self.CONFIG
        # in ProcessBlockData.
        config = self.config
        property_package = config.property_package

        if len(property_package.phase_list) != 1:
            raise ValueError(
                "%s can only be constructed with a "
                "single-phase property package.\n"
                "Got phases %s."
                % (self.__class__, [p for p in property_package.phase_list])
            )
        self.phase = next(iter(property_package.phase_list))
        if self.phase != "Vap":
            raise ValueError(
                '%s can only be constructed with a single phase, "Vap".'
                "Got phase %s." % (self.__class__, self.phase)
            )

        #
        # Make sure property package has the properties that this unit
        # model requires.
        #
        property_dict = property_package.get_metadata().properties
        if not property_dict["pressure"].supported:
            raise ValueError(
                "Property package supplied to pipeline must have a property "
                "for 'pressure', which was not found in %s." % type(property_package)
            )

        #
        # Set up control volume:
        #
        # Anything not valid for CV config will have to be removed here.
        cv_config = config()
        self.control_volume = ControlVolume1DBlock(**cv_config)
        self.control_volume.add_geometry()
        self.control_volume.add_state_blocks(
            has_phase_equilibrium=False,
        )
        self.control_volume.add_phase_component_balances()

        self.add_diameter()
        self.add_friction_factor()

        #
        # Add constraints indexed by space before applying transformation
        #
        self.add_flow_mass_linking_constraint()
        self.add_momentum_balance_equation()

        #
        # Apply spatial discretization transformation
        #
        self.control_volume.apply_transformation()

        #
        # Deactivate momentum balance at point where discretization equation
        # doesn't exist.
        #
        t0 = self.flowsheet().time.first()
        x0 = self.control_volume.length_domain.first()
        xf = self.control_volume.length_domain.last()
        dynamic = self.flowsheet().config.dynamic
        if dynamic:
            # In a dynamic model with an implicit time discretization,
            # at t0 and the point in space where the spatial discretization
            # equation is not defined, the momentum balance contains two
            # variables, the time and space derivatives, that are not present
            # in any other equations. For the model to have a perfect matching,
            # this equation must be deactivated.
            scheme = self.config.transformation_scheme
            momentum_bal = self.control_volume.momentum_balance
            if scheme in IMPLICIT_DISCRETIZATION_SCHEMES:
                momentum_bal[t0, x0].deactivate()
            elif scheme in EXPLICIT_DISCRETIZATION_SCHEMES:
                momentum_bal[t0, xf].deactivate()
            elif scheme in UNSUPPORTED_DISCRETIZATION_SCHEMES:
                raise ValueError(
                    "Discretization scheme %s is not supported for "
                    "dynamic pipelines" % scheme
                )

        # NOTE: Later we may try to support switches in direction of flow.
        # At this time these may need to be renamed...
        inlet_state = Reference(
            self.control_volume.properties[:, x0],
            ctype=StateBlock,
        )
        outlet_state = Reference(
            self.control_volume.properties[:, xf],
            ctype=StateBlock,
        )
        # I want my ports to be indexed only by time. The natural solution
        # is to build time-indexed state blocks called inlet_state and
        # outlet_state, build ports from these blocks, and connect their
        # state variables to my control volume state blocks with equality
        # constraints. I would rather not add these equality constraints,
        # however, so I use references.
        self.add_port(
            name="inlet_port",
            block=inlet_state,
            doc="The inlet to the pipeline",
        )
        self.add_port(
            name="outlet_port",
            block=outlet_state,
            doc="The outlet from the pipeline",
        )

        self.add_isothermal_constraint()

    def add_friction_factor(self, rugosity=None):
        if rugosity is None:
            rugosity = 0.025 * pyunits.mm
        rug_value = pyo_value(rugosity)
        rug_units = pyunits.get_units(rugosity)
        self.rugosity = Param(
            initialize=rug_value,
            units=rug_units,
            # doc=TODO,
            mutable=True,
        )

        diam_mm = pyunits.convert(self.diameter, pyunits.mm)
        friction_factor_expr = (2 * log10(3.7 * diam_mm / self.rugosity)) ** -2

        self.friction_factor = Var(
            initialize=pyo_value(friction_factor_expr),
            units=pyunits.dimensionless,
            doc="This is fraction factor lambda_l in V. Zavala's paper",
        )

        self.friction_factor_eqn = Constraint(
            expr=(self.friction_factor == friction_factor_expr)
        )

    def add_diameter(self):
        # We make diameter a variable as it may be more convenient
        # to fix diameter than to fix area.
        diam_units = pyunits.m
        self.diameter = Var(
            initialize=1.0,
            units=diam_units,
            doc="Diameter of the pipeline",
        )

        area = self.control_volume.area

        def diameter_rule(b):
            return (
                pyunits.convert(area, diam_units**2)
                == Constants.pi * self.diameter**2 / 4.0
            )

        self.diameter_eqn = Constraint(
            rule=diameter_rule,
            doc="Equation linking control volume area and pipeline diameter",
        )

    def add_flow_mass_linking_constraint(self):
        # We know flow_mass exists in the property package, but it might be
        # an expression. We need variables to construct the DerivativeVar.
        time = self.flowsheet().time
        space = self.control_volume.length_domain
        kghr = pyunits.kg / pyunits.hr
        cv = self.control_volume
        cv.flow_mass = Var(
            time, space, units=kghr, doc="Mass flow rate through the pipeline"
        )
        state = cv.properties

        def flow_mass_linking_rule(b, t, x):
            return b.flow_mass[t, x] - pyunits.convert(state[t, x].flow_mass, kghr) == 0

        cv.flow_mass_linking_constraint = Constraint(
            time, space, rule=flow_mass_linking_rule
        )

    def get_friction_term(self, t, x):
        """
        Equation (2.5b) in Zavala (2014)
        """
        cv = self.control_volume
        state = cv.properties[t, x]
        flow = state.flow_mass
        pressure = state.pressure
        diameter = self.diameter
        lam = self.friction_factor
        nu = state.speed_of_sound
        friction_term = (
            8
            * lam
            * nu**2
            / Constants.pi**2
            / diameter**5
            * flow
            * abs(flow)
            / pressure
        )
        return friction_term

    def add_flow_mass_dt(self):
        time = self.flowsheet().time
        # Should I attach this variable to the unit model or the
        # control volume?
        kghr2 = pyunits.kg / pyunits.hr**2
        cv = self.control_volume
        cv.flow_mass_dt = DerivativeVar(
            cv.flow_mass,
            units=kghr2,
            wrt=time,
        )

    def add_pressure_dx(self):
        # We add pressure_dx here rather than using the control volume
        # add_total_pressure_balances method because we don't want to add
        # a deltaP variable and pressure_balance equation.
        cv = self.control_volume
        cv.pressure = Reference(cv.properties[:, :].pressure)
        pressure_data = next(iter(cv.pressure.values()))
        space = cv.length_domain
        # Refrain from using metadata.derived_units here as I want the
        # units of my pressure variable (bar, hopefully) to override
        # the amalgamation of whatever base units I happen to use.
        p_units = pressure_data.get_units()
        cv.pressure_dx = DerivativeVar(
            cv.pressure,
            # Units are same as pressure because we assume length domain
            # is normalized.
            units=p_units,
            wrt=space,
        )

    def add_momentum_balance_equation(self):
        """
        Bulk momentum balance. This is Equation 2.5 in [1]
        """
        time = self.flowsheet().time
        dynamic = self.flowsheet().config.dynamic
        space = self.control_volume.length_domain
        if dynamic:
            self.add_flow_mass_dt()
        self.add_pressure_dx()
        cv = self.control_volume
        # TODO: Units of these equations should probably not
        # be hard-coded. These should be configurable.
        kgm2hr2 = pyunits.kg / pyunits.m**2 / pyunits.hr**2

        def momentum_balance_rule(b, t, x):
            # TODO: Should probably avoid having unit conversion calls
            # inside this function, as it will be called for every time/space
            # index.
            if dynamic:
                accum_expr = pyunits.convert(
                    cv.flow_mass_dt[t, x] / cv.area,
                    kgm2hr2,
                )
            else:
                accum_expr = 0.0
            flux_expr = pyunits.convert(cv.pressure_dx[t, x] / cv.length, kgm2hr2)
            friction_expr = pyunits.convert(self.get_friction_term(t, x), kgm2hr2)
            return accum_expr + flux_expr + friction_expr == 0

        cv.momentum_balance = Constraint(
            time,
            space,
            rule=momentum_balance_rule,
        )

    def add_isothermal_constraint(self):
        time = self.flowsheet().time
        length = self.control_volume.length_domain
        x0 = length.first()
        state = self.control_volume.properties

        def isothermal_rule(b, t, x):
            # NOTE: This constraint makes the pipeline non-templatizable
            # with respect to length. Also, we might want to skip at
            # xf depending on the direction of flow?
            if x != x0:
                x_prev = length.prev(x)
                return state[t, x].temperature == state[t, x_prev].temperature
            else:
                return Constraint.Skip

        self.state_isothermal_eqn = Constraint(time, length, rule=isothermal_rule)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Make sure the supercritical steam cycle example solves.
"""

__author__ = "Miguel Zamarripa"

import pytest
from pyomo.environ import check_optimal_termination, value
from pyomo.util.check_units import assert_units_consistent

import idaes.models_extra.power_generation.flowsheets.supercritical_power_plant.boiler_subflowsheet_build as blr
import idaes.models_extra.power_generation.flowsheets.supercritical_power_plant.SCPC_full_plant as SCPC
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.properties.general_helmholtz import helmholtz_available


@pytest.fixture(scope="module")
def boiler():
    m, solver = blr.main()

    m.solver = solver

    return m


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_init(boiler):
    # initialize each unit at the time
    blr.initialize(boiler)
    blr.unfix_inlets(boiler)
    # check that the model solved properly and has 0 degrees of freedom
    assert degrees_of_freedom(boiler) == 0


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_unit_consistency(boiler):
    assert_units_consistent(boiler)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_boiler(boiler):
    # unfix inlets to build arcs at the flowsheet level
    boiler.fs.ATMP1.outlet.enth_mol[0].fix(62710.01)
    boiler.fs.ATMP1.SprayWater.flow_mol[0].unfix()
    result = boiler.solver.solve(boiler, tee=False)
    assert check_optimal_termination(result)
    assert value(
        boiler.fs.ECON.cold_side.properties_out[0].temperature
    ) == pytest.approx(521.009, 1)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_power_plant():
    # SCPC.main imports and solves the SCPC Power Plant Flowsheet
    m, results = SCPC.main()
    assert check_optimal_termination(results)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Demonstration flowsheet for testing purposes.

Constructs a basic flowsheet for a benzene-toluene system with a mixer, heater
and flash unit.
"""
from pyomo.environ import ConcreteModel, TransformationFactory
from pyomo.network import Arc

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale

from idaes.core.util.initialization import propagate_state
from idaes.models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)
from idaes.models.unit_models import Mixer, Heater, Flash

import idaes.logger as idaeslog


def build_flowsheet():
    """Build demo flowsheet"""
    m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.BT_props = BTXParameterBlock()

    m.fs.M01 = Mixer(property_package=m.fs.BT_props)

    m.fs.H02 = Heater(property_package=m.fs.BT_props)

    m.fs.F03 = Flash(property_package=m.fs.BT_props)

    m.fs.s01 = Arc(source=m.fs.M01.outlet, destination=m.fs.H02.inlet)
    m.fs.s02 = Arc(source=m.fs.H02.outlet, destination=m.fs.F03.inlet)

    TransformationFactory("network.expand_arcs").apply_to(m.fs)

    return m


def set_scaling(m):
    """Set scaling for demo flowsheet"""

    iscale.set_scaling_factor(m.fs.M01.inlet_1_state[0].flow_mol_phase, 1e6)
    iscale.set_scaling_factor(m.fs.M01.inlet_1_state[0].flow_mol_phase["Liq"], 1e6)
    iscale.set_scaling_factor(m.fs.M01.inlet_1_state[0].flow_mol_phase["Vap"], 1)
    iscale.set_scaling_factor(
        m.fs.M01.inlet_1_state[0].mole_frac_phase_comp["Liq", "benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_1_state[0].mole_frac_phase_comp["Vap", "benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_1_state[0].mole_frac_phase_comp["Liq", "toluene"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_1_state[0].mole_frac_phase_comp["Vap", "toluene"], 1e6
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_1_state[0].pressure_sat_comp["benzene"], 1e-5
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_1_state[0].pressure_sat_comp["toluene"], 1e-4
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_1_state[0].enth_mol_phase_comp["Liq", "benzene"], 1e-4
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_1_state[0].enth_mol_phase_comp["Vap", "benzene"], 1e-4
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_1_state[0].enth_mol_phase_comp["Liq", "toluene"], 1e-4
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_1_state[0].enth_mol_phase_comp["Vap", "toluene"], 1e-4
    )

    iscale.set_scaling_factor(m.fs.M01.inlet_2_state[0].flow_mol_phase, 1e1)
    iscale.set_scaling_factor(m.fs.M01.inlet_2_state[0].flow_mol_phase["Vap"], 1e1)

    iscale.set_scaling_factor(
        m.fs.M01.inlet_2_state[0].mole_frac_phase_comp["Liq", "benzene"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_2_state[0].mole_frac_phase_comp["Vap", "benzene"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_2_state[0].mole_frac_phase_comp["Liq", "toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_2_state[0].mole_frac_phase_comp["Vap", "toluene"], 1
    )

    iscale.set_scaling_factor(
        m.fs.M01.inlet_2_state[0].pressure_sat_comp["benzene"], 1e-6
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_2_state[0].pressure_sat_comp["toluene"], 1e-6
    )

    iscale.set_scaling_factor(
        m.fs.M01.inlet_2_state[0].enth_mol_phase_comp["Liq", "benzene"], 1e-4
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_2_state[0].enth_mol_phase_comp["Vap", "benzene"], 1e-4
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_2_state[0].enth_mol_phase_comp["Liq", "toluene"], 1e-4
    )
    iscale.set_scaling_factor(
        m.fs.M01.inlet_2_state[0].enth_mol_phase_comp["Vap", "toluene"], 1e-4
    )

    iscale.set_scaling_factor(
        m.fs.M01.mixed_state[0].enth_mol_phase_comp["Liq", "benzene"], 1e-4
    )
    iscale.set_scaling_factor(
        m.fs.M01.mixed_state[0].enth_mol_phase_comp["Vap", "benzene"], 1e-4
    )
    iscale.set_scaling_factor(
        m.fs.M01.mixed_state[0].enth_mol_phase_comp["Liq", "toluene"], 1e-4
    )
    iscale.set_scaling_factor(
        m.fs.M01.mixed_state[0].enth_mol_phase_comp["Vap", "toluene"], 1e-4
    )
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].enthalpy_flow_terms["Liq"], 1)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].enthalpy_flow_terms["Vap"], 1)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].eq_total, 1)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].eq_sum_mol_frac, 1)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].eq_mol_frac_out, 1)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].eq_comp["benzene"], 1)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].eq_comp["toluene"], 1)
    iscale.set_scaling_factor(
        m.fs.M01.mixed_state[0].eq_phase_equilibrium["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.M01.mixed_state[0].eq_phase_equilibrium["toluene"], 1
    )
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].eq_P_vap["benzene"], 1)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].eq_P_vap["toluene"], 1)

    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].pressure, 1e-5)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].pressure_sat_comp, 1e-5)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].temperature, 1e-2)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].flow_mol, 1)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].flow_mol_phase, 1)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].mole_frac_comp["benzene"], 1e1)
    iscale.set_scaling_factor(m.fs.M01.mixed_state[0].mole_frac_comp["toluene"], 1e1)

    iscale.set_scaling_factor(m.fs.H02.control_volume.heat, 1e-5)
    iscale.set_scaling_factor(m.fs.H02.control_volume.properties_in[0].pressure, 1e-5)
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].pressure_sat_comp, 1e-5
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].temperature, 1e-2
    )
    iscale.set_scaling_factor(m.fs.H02.control_volume.properties_in[0].flow_mol, 1)
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].flow_mol_phase, 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].mole_frac_comp["benzene"], 1e1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].mole_frac_comp["toluene"], 1e1
    )
    iscale.set_scaling_factor(m.fs.H02.control_volume.properties_out[0].flow_mol, 1)
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].flow_mol_phase, 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].mole_frac_comp["benzene"], 1e1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].mole_frac_comp["toluene"], 1e1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].temperature, 1e-2
    )
    iscale.set_scaling_factor(m.fs.H02.control_volume.properties_out[0].pressure, 1e-5)
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].pressure_sat_comp, 1e-5
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].enth_mol_phase_comp["Liq", "benzene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].enth_mol_phase_comp["Vap", "benzene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].enth_mol_phase_comp["Liq", "toluene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].enth_mol_phase_comp["Vap", "toluene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].eq_comp["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].eq_comp["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].eq_comp["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].eq_comp["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].eq_phase_equilibrium["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0].eq_phase_equilibrium["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].eq_phase_equilibrium["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].eq_phase_equilibrium["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0.0].eq_P_vap["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0.0].eq_P_vap["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0.0].eq_P_vap["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0.0].eq_P_vap["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].enth_mol_phase_comp["Liq", "benzene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].enth_mol_phase_comp["Vap", "benzene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].enth_mol_phase_comp["Liq", "toluene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0].enth_mol_phase_comp["Vap", "toluene"],
        1e-4,
    )
    iscale.set_scaling_factor(m.fs.H02.control_volume.properties_in[0.0].eq_total, 1)
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0.0].eq_sum_mol_frac, 1
    )
    iscale.set_scaling_factor(m.fs.H02.control_volume.properties_out[0.0].eq_total, 1)
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0.0].eq_sum_mol_frac, 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_out[0.0].eq_mol_frac_out, 1
    )

    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0.0].material_flow_terms[
            "Liq", "benzene"
        ],
        1,
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0.0].material_flow_terms[
            "Vap", "benzene"
        ],
        1,
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0.0].material_flow_terms[
            "Liq", "toluene"
        ],
        1,
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0.0].material_flow_terms[
            "Vap", "toluene"
        ],
        1,
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0.0].enthalpy_flow_terms["Liq"], 1
    )
    iscale.set_scaling_factor(
        m.fs.H02.control_volume.properties_in[0.0].enthalpy_flow_terms["Vap"], 1
    )

    iscale.set_scaling_factor(m.fs.F03.control_volume.heat, 1)
    iscale.set_scaling_factor(m.fs.F03.control_volume.properties_in[0].pressure, 1e-5)
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].pressure_sat_comp, 1e-5
    )
    iscale.set_scaling_factor(m.fs.F03.control_volume.properties_out[0].pressure, 1e-5)
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0].pressure_sat_comp, 1e-5
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].temperature, 1e-2
    )
    iscale.set_scaling_factor(m.fs.F03.control_volume.properties_in[0].flow_mol, 1)
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].flow_mol_phase, 1
    )
    iscale.set_scaling_factor(m.fs.F03.control_volume.properties_out[0].flow_mol, 1)
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0].flow_mol_phase, 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].mole_frac_comp["benzene"], 1e1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].mole_frac_comp["toluene"], 1e1
    )

    iscale.set_scaling_factor(m.fs.F03.control_volume.properties_in[0.0].eq_total, 1)
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0.0].eq_sum_mol_frac, 1
    )
    iscale.set_scaling_factor(m.fs.F03.control_volume.properties_out[0.0].eq_total, 1)
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0.0].eq_sum_mol_frac, 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0.0].eq_mol_frac_out, 1
    )

    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].enth_mol_phase_comp["Liq", "benzene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].enth_mol_phase_comp["Vap", "benzene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].enth_mol_phase_comp["Liq", "toluene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].enth_mol_phase_comp["Vap", "toluene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0].enth_mol_phase_comp["Liq", "benzene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0].enth_mol_phase_comp["Vap", "benzene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0].enth_mol_phase_comp["Liq", "toluene"],
        1e-4,
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0].enth_mol_phase_comp["Vap", "toluene"],
        1e-4,
    )

    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].eq_comp["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].eq_comp["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0].eq_comp["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0].eq_comp["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].eq_phase_equilibrium["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0].eq_phase_equilibrium["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0].eq_phase_equilibrium["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0].eq_phase_equilibrium["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0.0].eq_P_vap["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0.0].eq_P_vap["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0.0].eq_P_vap["benzene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_out[0.0].eq_P_vap["toluene"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0.0].material_flow_terms[
            "Liq", "benzene"
        ],
        1,
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0.0].material_flow_terms[
            "Vap", "benzene"
        ],
        1,
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0.0].material_flow_terms[
            "Liq", "toluene"
        ],
        1,
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0.0].material_flow_terms[
            "Vap", "toluene"
        ],
        1,
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0.0].enthalpy_flow_terms["Liq"], 1
    )
    iscale.set_scaling_factor(
        m.fs.F03.control_volume.properties_in[0.0].enthalpy_flow_terms["Vap"], 1
    )

    iscale.calculate_scaling_factors(m)


def set_dof(m):
    """Set degrees of freedom for demo flowsheet"""
    eps = 1e-5
    m.fs.M01.inlet_1.flow_mol.fix(1.0)
    m.fs.M01.inlet_1.mole_frac_comp[:, "benzene"].fix(1 - eps)
    m.fs.M01.inlet_1.mole_frac_comp[:, "toluene"].fix(eps)
    m.fs.M01.inlet_1.pressure.fix(101325)
    m.fs.M01.inlet_1.temperature.fix(370)

    m.fs.M01.inlet_2.flow_mol.fix(1.0)
    m.fs.M01.inlet_2.mole_frac_comp[:, "benzene"].fix(eps)
    m.fs.M01.inlet_2.mole_frac_comp[:, "toluene"].fix(1 - eps)
    m.fs.M01.inlet_2.pressure.fix(1.3e5)
    m.fs.M01.inlet_2.temperature.fix(380)

    m.fs.H02.outlet.temperature[0].fix(370)

    m.fs.F03.heat_duty.fix(1e-6)
    m.fs.F03.deltaP.fix(1e-6)


def initialize_flowsheet(m):
    """Initialize demo flowsheet"""
    m.fs.M01.initialize(outlvl=idaeslog.WARNING)

    propagate_state(m.fs.s01)
    m.fs.H02.initialize(outlvl=idaeslog.WARNING)
    propagate_state(m.fs.s02)
    m.fs.F03.initialize(outlvl=idaeslog.WARNING)


def solve_flowsheet(m, stee=False):
    """Solve demo flowsheet"""
    solver = get_solver("ipopt_v2")
    solver.solve(m, tee=stee)


def display_results(m):
    """Display results for flowsheet"""
    m.fs.M01.outlet.display()
    m.fs.H02.outlet.display()
    m.fs.F03.vap_outlet.display()
    m.fs.F03.liq_outlet.display()


if __name__ == "__main__":
    m = build_flowsheet()

    set_scaling(m)

    set_dof(m)

    initialize_flowsheet(m)

    solve_flowsheet(m, stee=True)

    display_results(m)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Author: Andrew Lee
"""
import pytest
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Set,
    value,
    Var,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.unittest import assertStructuredAlmostEqual

from idaes.core import Component
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    fixed_variables_set,
    activated_constraints_set,
)
from idaes.core.solvers import get_solver

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE

from idaes.models.properties.modular_properties.examples.BT_ideal import configuration

from idaes.models.properties.tests.test_harness import PropertyTestHarness


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


def _as_quantity(x):
    unit = pyunits.get_units(x)
    if unit is None:
        unit = pyunits.dimensionless
    return value(x) * unit._get_pint_unit()


class TestBTIdeal(PropertyTestHarness):
    def configure(self):
        self.prop_pack = GenericParameterBlock
        self.param_args = configuration
        self.prop_args = {}
        self.has_density_terms = True


class TestParamBlock(object):
    @pytest.mark.unit
    def test_build(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration)

        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 2
        for i in model.params.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.params.Liq.is_liquid_phase()
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 2
        for i in model.params.component_list:
            assert i in ["benzene", "toluene"]
            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 4
        for i in model.params._phase_component_set:
            assert i in [
                ("Liq", "benzene"),
                ("Liq", "toluene"),
                ("Vap", "benzene"),
                ("Vap", "toluene"),
            ]

        assert model.params.config.state_definition == FTPx

        assertStructuredAlmostEqual(
            model.params.config.state_bounds,
            {
                "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
                "temperature": (273.15, 300, 450, pyunits.K),
                "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
            },
            item_callback=_as_quantity,
        )

        assert model.params.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE
        }

        assert isinstance(model.params.phase_equilibrium_idx, Set)
        assert len(model.params.phase_equilibrium_idx) == 2
        for i in model.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2"]

        assert model.params.phase_equilibrium_list == {
            "PE1": ["benzene", ("Vap", "Liq")],
            "PE2": ["toluene", ("Vap", "Liq")],
        }

        assert model.params.pressure_ref.value == 1e5
        assert model.params.temperature_ref.value == 300

        assert model.params.benzene.mw.value == 78.1136e-3
        assert model.params.benzene.pressure_crit.value == 48.9e5
        assert model.params.benzene.temperature_crit.value == 562.2

        assert model.params.toluene.mw.value == 92.1405e-3
        assert model.params.toluene.pressure_crit.value == 41e5
        assert model.params.toluene.temperature_crit.value == 591.8

        assert_units_consistent(model)


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration)

        model.props = model.params.build_state_block([1], defined_state=True)

        model.props[1].calculate_scaling_factors()

        # Fix state
        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(368)
        model.props[1].pressure.fix(101325)
        model.props[1].mole_frac_comp["benzene"].fix(0.5)
        model.props[1].mole_frac_comp["toluene"].fix(0.5)

        return model

    @pytest.mark.unit
    def test_build(self, model):
        # Check state variable values and bounds
        assert isinstance(model.props[1].flow_mol, Var)
        assert value(model.props[1].flow_mol) == 1
        assert model.props[1].flow_mol.ub == 1000
        assert model.props[1].flow_mol.lb == 0

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 101325
        assert model.props[1].pressure.ub == 1e6
        assert model.props[1].pressure.lb == 5e4

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 368
        assert model.props[1].temperature.ub == 450
        assert model.props[1].temperature.lb == 273.15

        assert isinstance(model.props[1].mole_frac_comp, Var)
        assert len(model.props[1].mole_frac_comp) == 2
        for i in model.props[1].mole_frac_comp:
            assert value(model.props[1].mole_frac_comp[i]) == 0.5

        assert_units_consistent(model)

    @pytest.mark.unit
    def test_define_state_vars(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol", "mole_frac_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_port_members(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol", "mole_frac_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_display_vars(self, model):
        sv = model.props[1].define_display_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in [
                "Total Molar Flowrate",
                "Total Mole Fraction",
                "Temperature",
                "Pressure",
            ]

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model.props[1]) == 0

    @pytest.mark.unit
    def test_basic_scaling(self, model):
        model.props[1].scaling_factor.display()
        assert len(model.props[1].scaling_factor) == 23
        assert (
            model.props[1].scaling_factor[
                model.props[1]._mole_frac_tbub["Vap", "Liq", "benzene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1]._mole_frac_tbub["Vap", "Liq", "toluene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1]._mole_frac_tdew["Vap", "Liq", "benzene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1]._mole_frac_tdew["Vap", "Liq", "toluene"]
            ]
            == 1000
        )
        assert model.props[1].scaling_factor[model.props[1]._t1_Vap_Liq] == 1e-2
        assert model.props[1].scaling_factor[model.props[1]._teq["Vap", "Liq"]] == 1e-2
        assert model.props[1].scaling_factor[model.props[1].flow_mol] == 1e-2
        assert (
            model.props[1].scaling_factor[model.props[1].flow_mol_phase["Liq"]] == 1e-2
        )
        assert (
            model.props[1].scaling_factor[model.props[1].flow_mol_phase["Vap"]] == 1e-2
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].flow_mol_phase_comp["Liq", "benzene"]
            ]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].flow_mol_phase_comp["Liq", "toluene"]
            ]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].flow_mol_phase_comp["Vap", "benzene"]
            ]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].flow_mol_phase_comp["Vap", "toluene"]
            ]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[model.props[1].mole_frac_comp["benzene"]]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[model.props[1].mole_frac_comp["toluene"]]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].mole_frac_phase_comp["Liq", "benzene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].mole_frac_phase_comp["Liq", "toluene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].mole_frac_phase_comp["Vap", "benzene"]
            ]
            == 1000
        )
        assert (
            model.props[1].scaling_factor[
                model.props[1].mole_frac_phase_comp["Vap", "toluene"]
            ]
            == 1000
        )
        assert model.props[1].scaling_factor[model.props[1].pressure] == 1e-5
        assert model.props[1].scaling_factor[model.props[1].temperature] == 1e-2
        assert (
            model.props[1].scaling_factor[
                model.props[1].temperature_bubble["Vap", "Liq"]
            ]
            == 1e-2
        )
        assert (
            model.props[1].scaling_factor[model.props[1].temperature_dew["Vap", "Liq"]]
            == 1e-2
        )

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6})

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        # Check phase equilibrium results
        assert model.props[1].mole_frac_phase_comp[
            "Liq", "benzene"
        ].value == pytest.approx(0.4121, abs=1e-4)
        assert model.props[1].mole_frac_phase_comp[
            "Vap", "benzene"
        ].value == pytest.approx(0.6339, abs=1e-4)
        assert model.props[1].phase_frac["Vap"].value == pytest.approx(0.3961, abs=1e-4)

        assert value(
            model.props[1].conc_mol_phase_comp["Vap", "benzene"]
        ) == pytest.approx(20.9946, abs=1e-4)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, model):
        model.props[1].report()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for CLC gas phase thermo state block; tests for construction and solve
Author: Chinedu Okoli
"""

import pytest

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Var,
    value,
    units as pyunits,
)
from pyomo.contrib.incidence_analysis import (
    solve_strongly_connected_components,
)
from pyomo.common.collections import ComponentMap
import pyomo.common.unittest as unittest
from pyomo.common.unittest import TestCase
from pyomo.util.subsystems import ParamSweeper
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_large_residuals,
    fixed_variables_set,
    activated_constraints_set,
)

from idaes.core.solvers import get_solver
from idaes.core.util import scaling as iscale

from idaes.models_extra.gas_solid_contactors.properties.oxygen_iron_OC_oxidation.gas_phase_thermo import (
    GasPhaseParameterBlock,
)

# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.fixture(scope="class")
def gas_prop():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # gas properties and state inlet block
    m.fs.properties = GasPhaseParameterBlock()
    m.fs.unit = m.fs.properties.build_state_block(
        [0], parameters=m.fs.properties, defined_state=True
    )

    m.fs.unit[0].flow_mol.fix(1)
    m.fs.unit[0].temperature.fix(450)
    m.fs.unit[0].pressure.fix(1.60e5)
    m.fs.unit[0].mole_frac_comp["CO2"].fix(0.0004)
    m.fs.unit[0].mole_frac_comp["H2O"].fix(0.0093)
    m.fs.unit[0].mole_frac_comp["O2"].fix(0.2095)
    m.fs.unit[0].mole_frac_comp["N2"].fix(0.7808)

    return m


@pytest.mark.unit
def test_build_inlet_state_block(gas_prop):
    assert isinstance(gas_prop.fs.unit[0].mw, Var)
    assert isinstance(gas_prop.fs.unit[0].dens_mol, Var)
    assert isinstance(gas_prop.fs.unit[0].dens_mol_comp, Var)
    assert isinstance(gas_prop.fs.unit[0].dens_mass, Var)
    assert isinstance(gas_prop.fs.unit[0].cp_mol_comp, Var)
    assert isinstance(gas_prop.fs.unit[0].cp_mol, Var)
    assert isinstance(gas_prop.fs.unit[0].cp_mass, Var)
    assert isinstance(gas_prop.fs.unit[0].visc_d, Var)


@pytest.mark.unit
def test_setInputs_state_block(gas_prop):
    assert degrees_of_freedom(gas_prop.fs.unit[0]) == 0


@pytest.fixture(scope="class")
def gas_prop_unscaled():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    # gas properties and state inlet block
    m.fs.properties = GasPhaseParameterBlock()
    m.fs.unit = m.fs.properties.build_state_block(
        [0], parameters=m.fs.properties, defined_state=True
    )

    m.fs.unit[0].flow_mol.fix(1)
    m.fs.unit[0].temperature.fix(450)
    m.fs.unit[0].pressure.fix(1.60e5)
    m.fs.unit[0].mole_frac_comp["CO2"].fix(0.0004)
    m.fs.unit[0].mole_frac_comp["H2O"].fix(0.0093)
    m.fs.unit[0].mole_frac_comp["O2"].fix(0.2095)
    m.fs.unit[0].mole_frac_comp["N2"].fix(0.7808)

    return m


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_unscaled(gas_prop_unscaled):
    orig_fixed_vars = fixed_variables_set(gas_prop_unscaled)
    orig_act_consts = activated_constraints_set(gas_prop_unscaled)

    gas_prop_unscaled.fs.unit.initialize()

    assert degrees_of_freedom(gas_prop_unscaled) == 0

    fin_fixed_vars = fixed_variables_set(gas_prop_unscaled)
    fin_act_consts = activated_constraints_set(gas_prop_unscaled)

    assert len(fin_act_consts) == len(orig_act_consts)
    assert len(fin_fixed_vars) == len(orig_fixed_vars)

    for c in fin_act_consts:
        assert c in orig_act_consts
    for v in fin_fixed_vars:
        assert v in orig_fixed_vars


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve_unscaled(gas_prop_unscaled):

    assert hasattr(gas_prop_unscaled.fs.unit[0], "mw")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "dens_mol")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "dens_mol_comp")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "dens_mass")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "visc_d")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "therm_cond")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "diffus_comp")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "cp_mol_comp")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "cp_mol")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "cp_mass")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "enth_mol")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "enth_mol_comp")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "entr_mol")

    assert hasattr(gas_prop_unscaled.fs.unit[0], "get_material_flow_terms")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "get_enthalpy_flow_terms")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "get_material_density_terms")
    assert hasattr(gas_prop_unscaled.fs.unit[0], "get_energy_density_terms")

    results = solver.solve(gas_prop_unscaled)

    # Check for optimal solution
    assert check_optimal_termination(results)


@pytest.mark.component
def test_scaling(gas_prop):
    # Calculate scaling factors

    # Construct property methods to build the constraints
    assert hasattr(gas_prop.fs.unit[0], "mw")
    assert hasattr(gas_prop.fs.unit[0], "dens_mol")
    assert hasattr(gas_prop.fs.unit[0], "dens_mol_comp")
    assert hasattr(gas_prop.fs.unit[0], "dens_mass")
    assert hasattr(gas_prop.fs.unit[0], "visc_d")
    assert hasattr(gas_prop.fs.unit[0], "therm_cond")
    assert hasattr(gas_prop.fs.unit[0], "diffus_comp")
    assert hasattr(gas_prop.fs.unit[0], "cp_mol_comp")
    assert hasattr(gas_prop.fs.unit[0], "cp_mol")
    assert hasattr(gas_prop.fs.unit[0], "cp_mass")
    assert hasattr(gas_prop.fs.unit[0], "enth_mol")
    assert hasattr(gas_prop.fs.unit[0], "enth_mol_comp")
    assert hasattr(gas_prop.fs.unit[0], "entr_mol")

    # Call flow and density methods to construct flow and density expressions
    for i in gas_prop.fs.unit[0]._params.component_list:
        gas_prop.fs.unit[0].get_material_flow_terms("Vap", i)
        gas_prop.fs.unit[0].get_material_density_terms("Vap", i)
    gas_prop.fs.unit[0].get_enthalpy_flow_terms("Vap")
    gas_prop.fs.unit[0].get_energy_density_terms("Vap")

    # Calculate scaling factors now that constraints/expressions are built
    iscale.calculate_scaling_factors(gas_prop)

    # Test scaling
    assert pytest.approx(1e-3, rel=1e-5) == iscale.get_scaling_factor(
        gas_prop.fs.unit[0].dens_mol
    )

    for i, c in gas_prop.fs.unit[0].material_flow_terms.items():
        assert pytest.approx(1e-2, rel=1e-5) == iscale.get_scaling_factor(c)
    for i, c in gas_prop.fs.unit[0].material_density_terms.items():
        assert pytest.approx(1e-2, rel=1e-5) == iscale.get_scaling_factor(c)
    for i, c in gas_prop.fs.unit[0].energy_density_terms.items():
        assert pytest.approx(1e-9, rel=1e-5) == iscale.get_scaling_factor(c)
    for i, c in gas_prop.fs.unit[0].enthalpy_flow_terms.items():
        assert pytest.approx(1e-9, rel=1e-5) == iscale.get_scaling_factor(c)

    assert pytest.approx(
        1e2, rel=1e-5
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        gas_prop.fs.unit[0].mw_eqn
    )
    assert pytest.approx(
        1e-5, rel=1e-5
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        gas_prop.fs.unit[0].ideal_gas
    )
    for i, c in gas_prop.fs.unit[0].comp_conc_eqn.items():
        assert pytest.approx(
            1e-2, rel=1e-5
        ) == iscale.get_constraint_transform_applied_scaling_factor(c)
    for i, c in gas_prop.fs.unit[0].visc_d_constraint.items():
        assert pytest.approx(
            1e5, rel=1e-5
        ) == iscale.get_constraint_transform_applied_scaling_factor(c)
    for i, c in gas_prop.fs.unit[0].diffus_comp_constraint.items():
        assert pytest.approx(
            1e5, rel=1e-5
        ) == iscale.get_constraint_transform_applied_scaling_factor(c)
    assert pytest.approx(
        1, rel=1e-5
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        gas_prop.fs.unit[0].therm_cond_constraint
    )
    for i, c in gas_prop.fs.unit[0].cp_shomate_eqn.items():
        assert pytest.approx(
            1e-6, rel=1e-5
        ) == iscale.get_constraint_transform_applied_scaling_factor(c)
    assert pytest.approx(
        1e-6, rel=1e-5
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        gas_prop.fs.unit[0].mixture_heat_capacity_eqn
    )
    assert pytest.approx(
        1e-6, rel=1e-5
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        gas_prop.fs.unit[0].cp_mass_basis
    )
    for i, c in gas_prop.fs.unit[0].enthalpy_shomate_eqn.items():
        assert pytest.approx(
            1e-6, rel=1e-5
        ) == iscale.get_constraint_transform_applied_scaling_factor(c)
    assert pytest.approx(
        1e-6, rel=1e-5
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        gas_prop.fs.unit[0].mixture_enthalpy_eqn
    )
    assert pytest.approx(
        1e-5, rel=1e-5
    ) == iscale.get_constraint_transform_applied_scaling_factor(
        gas_prop.fs.unit[0].entropy_correlation
    )


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize(gas_prop):
    orig_fixed_vars = fixed_variables_set(gas_prop)
    orig_act_consts = activated_constraints_set(gas_prop)

    gas_prop.fs.unit.initialize()

    assert degrees_of_freedom(gas_prop) == 0

    fin_fixed_vars = fixed_variables_set(gas_prop)
    fin_act_consts = activated_constraints_set(gas_prop)

    assert len(fin_act_consts) == len(orig_act_consts)
    assert len(fin_fixed_vars) == len(orig_fixed_vars)

    for c in fin_act_consts:
        assert c in orig_act_consts
    for v in fin_fixed_vars:
        assert v in orig_fixed_vars


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve(gas_prop):

    assert hasattr(gas_prop.fs.unit[0], "mw")
    assert hasattr(gas_prop.fs.unit[0], "cp_mol")
    assert hasattr(gas_prop.fs.unit[0], "enth_mol")

    results = solver.solve(gas_prop)

    # Check for optimal solution
    assert check_optimal_termination(results)


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solution(gas_prop):
    assert pytest.approx(1, rel=1e-5) == gas_prop.fs.unit[0].mw.value
    assert pytest.approx(1, rel=1e-5) == gas_prop.fs.unit[0].cp_mol.value
    assert pytest.approx(1, rel=1e-5) == gas_prop.fs.unit[0].enth_mol.value


@pytest.mark.component
def test_units_consistent(gas_prop):

    # Construct property methods to build the constraints
    assert hasattr(gas_prop.fs.unit[0], "mw")
    assert hasattr(gas_prop.fs.unit[0], "dens_mol")
    assert hasattr(gas_prop.fs.unit[0], "dens_mol_comp")
    assert hasattr(gas_prop.fs.unit[0], "dens_mass")
    assert hasattr(gas_prop.fs.unit[0], "visc_d")
    assert hasattr(gas_prop.fs.unit[0], "therm_cond")
    assert hasattr(gas_prop.fs.unit[0], "diffus_comp")
    assert hasattr(gas_prop.fs.unit[0], "cp_mol_comp")
    assert hasattr(gas_prop.fs.unit[0], "cp_mol")
    assert hasattr(gas_prop.fs.unit[0], "cp_mass")
    assert hasattr(gas_prop.fs.unit[0], "enth_mol")
    assert hasattr(gas_prop.fs.unit[0], "enth_mol_comp")
    assert hasattr(gas_prop.fs.unit[0], "entr_mol")

    # Call flow and density methods to construct flow and density expressions
    for i in gas_prop.fs.unit[0]._params.component_list:
        gas_prop.fs.unit[0].get_material_flow_terms("Vap", i)
        gas_prop.fs.unit[0].get_material_density_terms("Vap", i)
    gas_prop.fs.unit[0].get_enthalpy_flow_terms("Vap")
    gas_prop.fs.unit[0].get_energy_density_terms("Vap")

    assert_units_consistent(gas_prop)


# -----------------------------------------------------------------------------


@pytest.mark.unit
def test_state_vars():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = GasPhaseParameterBlock()
    m.fs.state = m.fs.properties.build_state_block()

    assert isinstance(m.fs.state.flow_mol, Var)
    assert isinstance(m.fs.state.temperature, Var)
    assert isinstance(m.fs.state.pressure, Var)
    assert isinstance(m.fs.state.mole_frac_comp, Var)

    assert isinstance(m.fs.state.sum_component_eqn, Constraint)

    assert len(list(m.component_data_objects(Var))) == 7
    assert len(list(m.component_data_objects(Constraint))) == 1

    for name, var in m.fs.state.define_state_vars().items():
        # State vars should be included in the metadata with no method
        meta = m.fs.properties._metadata._properties[name]
        assert meta.method is None
        assert meta.supported


@pytest.mark.unit
def test_indexed_state_block():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = GasPhaseParameterBlock()
    m.fs.state = m.fs.properties.build_state_block([1, 2, 3])

    assert len(list(m.component_data_objects(Var))) == 21
    assert len(list(m.component_data_objects(Constraint))) == 3

    for i, state in m.fs.state.items():
        assert isinstance(state.flow_mol, Var)
        assert isinstance(state.temperature, Var)
        assert isinstance(state.pressure, Var)
        assert isinstance(state.mole_frac_comp, Var)

        assert isinstance(state.sum_component_eqn, Constraint)

        for name, var in state.define_state_vars().items():
            # State vars should be included in the metadata with no method
            meta = m.fs.properties._metadata._properties[name]
            assert meta.method is None
            assert meta.supported


@pytest.mark.unit
def test_property_construction_ordered():
    """
    The purpose of this test is to make sure each property can be
    constructed properly. Further, because we know that these
    variables form a DAG, we check that the addition of each new
    property adds one variable and constraint, i.e. that the
    matching defined in the parameter block is in topological order.
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = GasPhaseParameterBlock()
    m.fs.state = m.fs.properties.build_state_block()

    nvar = 7
    ncon = 1
    assert len(list(m.component_data_objects(Var))) == nvar
    assert len(list(m.component_data_objects(Constraint))) == ncon

    matching = [
        ("mw", "mw_eqn"),
        ("dens_mol", "ideal_gas"),
        ("dens_mol_comp", "comp_conc_eqn"),
        ("dens_mass", "dens_mass_basis"),
        ("visc_d", "visc_d_constraint"),
        ("diffus_comp", "diffus_comp_constraint"),
        ("therm_cond", "therm_cond_constraint"),
        ("cp_mol_comp", "cp_shomate_eqn"),
        ("cp_mol", "mixture_heat_capacity_eqn"),
        ("cp_mass", "cp_mass_basis"),
        ("enth_mol_comp", "enthalpy_shomate_eqn"),
        ("enth_mol", "mixture_enthalpy_eqn"),
        ("entr_mol", "entropy_correlation"),
    ]

    # Make sure the matching captures all the non-state vars.
    state_vars = m.fs.state.define_state_vars()
    n_state_vars = len(state_vars)
    n_vars = len(m.fs.properties._metadata.properties.list_supported_properties())
    assert len(matching) == n_vars - n_state_vars

    # Make sure we can add constraints one at a time, and only
    # add one additional variable. This is specific to this
    # property package.
    for varname, conname in matching:
        assert varname not in state_vars
        var = getattr(m.fs.state, varname)
        con = getattr(m.fs.state, conname)
        dim = len(var)
        nvar += dim
        ncon += dim
        assert dim == len(con)
        assert isinstance(var, Var)
        assert isinstance(con, Constraint)
        assert len(list(m.component_data_objects(Var))) == nvar
        assert len(list(m.component_data_objects(Constraint))) == ncon


@pytest.mark.unit
def test_property_construction_unordered():
    """
    The purpose of this test is to make sure "chained construction"
    of properties happens properly. I.e. if we add a property that
    requires several other properties, these are added as well.
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = GasPhaseParameterBlock()
    m.fs.state = m.fs.properties.build_state_block()

    nvar = 7
    ncon = 1
    assert len(list(m.component_data_objects(Var))) == nvar
    assert len(list(m.component_data_objects(Constraint))) == ncon

    # Access this property to force the construction of prerequisite
    # properties.
    m.fs.state.cp_mass

    # Make sure dependent properties have been properly added.
    assert hasattr(m.fs.state, "cp_mol")
    assert hasattr(m.fs.state, "cp_mol_comp")
    assert hasattr(m.fs.state, "mw")
    nvar += len(m.fs.state.cp_mol_comp)
    nvar += len(m.fs.state.cp_mol)
    nvar += len(m.fs.state.cp_mass)
    nvar += len(m.fs.state.mw)
    ncon += len(m.fs.state.cp_mol_comp)
    ncon += len(m.fs.state.cp_mol)
    ncon += len(m.fs.state.cp_mass)
    ncon += len(m.fs.state.mw)
    assert len(list(m.component_data_objects(Var))) == nvar
    assert len(list(m.component_data_objects(Constraint))) == ncon


@pytest.mark.unit
class TestProperties(TestCase):
    """
    The purpose of these tests is to ensure that the property package's
    state block can be used to solve for the specified properties, and
    that the values produced are what we expect. Some of these values are
    computed by hand, while others (viscosity, enthalpy, conductivity, etc.)
    are computed from these constraints we are testing.
    These tests pretty much just make sure the calculated values don't change.
    If the property equations change, and the values change as a result,
    the tolerances may be relaxed. If the new values are significantly
    different, this should be investigated.
    """

    def _make_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = GasPhaseParameterBlock()
        m.fs.state = m.fs.properties.build_state_block(defined_state=True)
        for var in m.fs.state.define_state_vars().values():
            var.fix()
        return m

    def test_mw(self):
        m = self._make_model()
        state = m.fs.state

        # Define a somewhat arbitrary set of values we'd like to use to
        # test molecular weight.
        n_scenario = 7
        state_values = {
            "flow_mol": [1.0 * pyunits.mol / pyunits.s] * n_scenario,
            "temperature": [300.0 * pyunits.K] * n_scenario,
            "pressure": [1.0 * pyunits.bar] * n_scenario,
            "mole_frac_comp[O2]": [1.0, 0.5, 0.25, 0.0, 0.0, 0.0, 0.0],
            "mole_frac_comp[N2]": [0.0, 0.5, 0.25, 1.0, 0.0, 0.0, 0.0],
            "mole_frac_comp[H2O]": [0.0, 0.0, 0.25, 0.0, 1.0, 0.0, 0.5],
            "mole_frac_comp[CO2]": [0.0, 0.0, 0.25, 0.0, 0.0, 1.0, 0.5],
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )
        target_values = {
            "mw": [
                32.0 * pyunits.g / pyunits.mol,
                30.0 * pyunits.g / pyunits.mol,
                30.5 * pyunits.g / pyunits.mol,
                28.0 * pyunits.g / pyunits.mol,
                18.0 * pyunits.g / pyunits.mol,
                44.0 * pyunits.g / pyunits.mol,
                31.0 * pyunits.g / pyunits.mol,
            ]
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        # Construct mw and all prerequisites
        state.mw

        param_sweeper = ParamSweeper(
            n_scenario,
            state_values,
            output_values=target_values,
        )
        with param_sweeper:
            for inputs, target in param_sweeper:
                solve_strongly_connected_components(state)

                # Check that state block has been been solved correctly
                assert number_large_residuals(state, tol=1e-8) == 0

                # Sanity check that inputs have been set properly
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(val, abs=1e-8)

                # Check that the state block computes the property values
                # we expect
                for var, val in target.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(val, abs=1e-8)

    def test_dens_mol(self):
        m = self._make_model()
        state = m.fs.state

        n_scen = 8
        state_values = {
            "flow_mol": [1.0 * pyunits.mol / pyunits.s] * n_scen,
            "temperature": [
                200.0 * pyunits.K,
                300.0 * pyunits.K,
                600.0 * pyunits.K,
                900.0 * pyunits.K,
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
            ],
            "pressure": [
                1.0 * pyunits.bar,
                1.0 * pyunits.bar,
                1.0 * pyunits.bar,
                1.0 * pyunits.bar,
                1.0 * pyunits.bar,
                0.5 * pyunits.bar,
                1.5 * pyunits.bar,
                2.0 * pyunits.bar,
            ],
            "mole_frac_comp[O2]": [0.25] * n_scen,
            "mole_frac_comp[N2]": [0.25] * n_scen,
            "mole_frac_comp[H2O]": [0.25] * n_scen,
            "mole_frac_comp[CO2]": [0.25] * n_scen,
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )

        u = pyunits.mol / pyunits.m**3
        target_values = {
            "dens_mol": [
                # Calculated by P*1e5/(T*8.314462618)
                60.136 * u,
                40.091 * u,
                20.045 * u,
                13.364 * u,
                10.023 * u,
                5.011 * u,
                15.034 * u,
                20.045 * u,
            ]
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        # Construct dens_mol and all prerequisites
        state.dens_mol
        assert_units_consistent(state.ideal_gas)

        param_sweeper = ParamSweeper(n_scen, state_values, output_values=target_values)
        with param_sweeper:
            for inputs, target in param_sweeper:
                solve_strongly_connected_components(state)

                # Check that state block equations have been converged
                assert number_large_residuals(state, tol=1e-8) == 0

                # Sanity check that inputs are what we expect
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(val, rel=1e-3)

                # Check that state block performs the calculations we expect
                for var, val in target.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(val, rel=1e-3)

    def test_dens_mol_comp(self):
        m = self._make_model()
        state = m.fs.state

        n_scen = 5
        state_values = {
            "flow_mol": [1.0 * pyunits.mol / pyunits.s] * n_scen,
            "temperature": [1200.0 * pyunits.K] * n_scen,
            "pressure": [1.0 * pyunits.bar] * n_scen,
            "mole_frac_comp[O2]": [1.0, 0.0, 0.0, 0.0, 0.25],
            "mole_frac_comp[N2]": [0.0, 1.0, 0.0, 0.0, 0.25],
            "mole_frac_comp[H2O]": [0.0, 0.0, 1.0, 0.0, 0.25],
            "mole_frac_comp[CO2]": [0.0, 0.0, 0.0, 1.0, 0.25],
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )

        u = pyunits.mol / pyunits.m**3
        target_values = {
            "dens_mol": [10.023 * u] * n_scen,
            "dens_mol_comp[O2]": [
                10.023 * u,
                0.0 * u,
                0.0 * u,
                0.0 * u,
                0.25 * 10.023 * u,
            ],
            "dens_mol_comp[N2]": [
                0.0 * u,
                10.023 * u,
                0.0 * u,
                0.0 * u,
                0.25 * 10.023 * u,
            ],
            "dens_mol_comp[H2O]": [
                0.0 * u,
                0.0 * u,
                10.023 * u,
                0.0 * u,
                0.25 * 10.023 * u,
            ],
            "dens_mol_comp[CO2]": [
                0.0 * u,
                0.0 * u,
                0.0 * u,
                10.023 * u,
                0.25 * 10.023 * u,
            ],
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        # Construct dens_mol_comp and all prerequisites
        state.dens_mol_comp

        param_sweeper = ParamSweeper(n_scen, state_values, output_values=target_values)
        with param_sweeper:
            for inputs, target in param_sweeper:
                solve_strongly_connected_components(state)

                # Make sure property equations have been converged
                assert number_large_residuals(state, tol=1e-8) == 0

                # Sanity check that inputs have been set properly
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    # ^ Problem converting units when value is zero
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                # Make sure property values are what we expect
                for var, val in target.items():
                    # val = value(pyunits.convert(val, var.get_units()))
                    # ^ Problem converting units when value is zero
                    assert var.value == pytest.approx(value(val), rel=1e-3)

    def test_visc_d(self):
        m = self._make_model()
        state = m.fs.state

        n_scen = 8
        state_values = {
            "flow_mol": [1.0 * pyunits.mol / pyunits.s] * n_scen,
            "temperature": [
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                300.0 * pyunits.K,
                600.0 * pyunits.K,
                900.0 * pyunits.K,
                1200.0 * pyunits.K,
            ],
            "pressure": [1.0 * pyunits.bar] * n_scen,
            "mole_frac_comp[O2]": [1.0, 0.0, 0.0, 0.0, 0.25, 0.25, 0.25, 0.25],
            "mole_frac_comp[N2]": [0.0, 1.0, 0.0, 0.0, 0.25, 0.25, 0.25, 0.25],
            "mole_frac_comp[H2O]": [0.0, 0.0, 1.0, 0.0, 0.25, 0.25, 0.25, 0.25],
            "mole_frac_comp[CO2]": [0.0, 0.0, 0.0, 1.0, 0.25, 0.25, 0.25, 0.25],
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )

        target_values = {
            "visc_d": [
                # These values were copied after a solve.
                # TODO: Cross-reference with another source.
                5.534440949133228e-05 * pyunits.Pa * pyunits.s,
                4.67667824296429e-05 * pyunits.Pa * pyunits.s,
                4.6232771210527155e-05 * pyunits.Pa * pyunits.s,
                4.512867970060493e-05 * pyunits.Pa * pyunits.s,
                1.6181534595116313e-05 * pyunits.Pa * pyunits.s,
                2.866222939903063e-05 * pyunits.Pa * pyunits.s,
                3.909320395131273e-05 * pyunits.Pa * pyunits.s,
                4.838841106600266e-05 * pyunits.Pa * pyunits.s,
            ]
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        # Construct visc_d and all prerequisites
        state.visc_d

        param_sweeper = ParamSweeper(n_scen, state_values, output_values=target_values)
        with param_sweeper:
            for inputs, target in param_sweeper:
                solve_strongly_connected_components(state)

                # Make sure property equations have been converged
                assert number_large_residuals(state, tol=1e-8) == 0

                # Sanity check that inputs are properly set
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                # Make sure properties have been calculated as expected
                for var, val in target.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

    def test_diffus_comp(self):
        m = self._make_model()
        state = m.fs.state

        n_scen = 11
        bar = pyunits.bar
        K = pyunits.K
        state_values = {
            "flow_mol": [1.0 * pyunits.mol / pyunits.s] * n_scen,
            "temperature": [
                1200.0 * K,
                1200.0 * K,
                1200.0 * K,
                1200.0 * K,
                300.0 * K,
                600.0 * K,
                900.0 * K,
                1200.0 * K,
                1200.0 * K,
                1200.0 * K,
                1200.0 * K,
            ],
            "pressure": [
                1.0 * bar,
                1.0 * bar,
                1.0 * bar,
                1.0 * bar,
                1.0 * bar,
                1.0 * bar,
                1.0 * bar,
                1.0 * bar,
                0.5 * bar,
                1.5 * bar,
                2.0 * bar,
            ],
            "mole_frac_comp[O2]": [
                # Note that diffusivity is not defined for a pure
                # component in itself (zero gradient, zero net diffusion)
                0.90,
                0.025,
                0.025,
                0.025,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
            ],
            "mole_frac_comp[N2]": [
                0.025,
                0.90,
                0.025,
                0.025,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
            ],
            "mole_frac_comp[H2O]": [
                0.025,
                0.025,
                0.90,
                0.025,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
            ],
            "mole_frac_comp[CO2]": [
                0.025,
                0.025,
                0.025,
                0.90,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
                0.25,
            ],
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )

        cm = pyunits.cm
        s = pyunits.s
        target_values = {
            # These values look reasonable
            # TODO: Verify with external source.
            "diffus_comp[O2]": [
                3.11792621830951e-5 * cm**2 / s,
                2.456227751888218e-5 * cm**2 / s,
                3.034740091620132e-5 * cm**2 / s,
                1.9479388404541838e-5 * cm**2 / s,
                0.206691259894311e-5 * cm**2 / s,
                0.6952237580376006e-5 * cm**2 / s,
                1.4134625566198689e-5 * cm**2 / s,
                2.3384446637321363e-5 * cm**2 / s,
                4.676889327464272e-5 * cm**2 / s,
                1.5589631091547582e-5 * cm**2 / s,
                1.1692223318660684e-5 * cm**2 / s,
            ],
            "diffus_comp[N2]": [
                2.457518754629481e-5 * cm**2 / s,
                3.1383309495574956e-5 * cm**2 / s,
                3.0334118458311523e-5 * cm**2 / s,
                1.9790010245966736e-5 * cm**2 / s,
                0.2080439152537244e-5 * cm**2 / s,
                0.6997735302088169e-5 * cm**2 / s,
                1.422712718932098e-5 * cm**2 / s,
                2.353748212168125e-5 * cm**2 / s,
                4.70749642433625e-5 * cm**2 / s,
                1.5691654747787498e-5 * cm**2 / s,
                1.176874106084062e-5 * cm**2 / s,
            ],
            "diffus_comp[H2O]": [
                3.0845168350215713e-5 * cm**2 / s,
                3.0811129521516416e-5 * cm**2 / s,
                3.719143107603082e-5 * cm**2 / s,
                2.5051274838003996e-5 * cm**2 / s,
                0.24654668546150185e-5 * cm**2 / s,
                0.8292808959890481e-5 * cm**2 / s,
                1.6860147281349098e-5 * cm**2 / s,
                2.7893573307023156e-5 * cm**2 / s,
                5.578714661404631e-5 * cm**2 / s,
                1.8595715538015434e-5 * cm**2 / s,
                1.3946786653511578e-5 * cm**2 / s,
            ],
            "diffus_comp[CO2]": [
                1.929322600571961e-5 * cm**2 / s,
                1.9589681584041365e-5 * cm**2 / s,
                2.4422863072776697e-5 * cm**2 / s,
                2.7098612589802444e-5 * cm**2 / s,
                0.17964011927809195e-5 * cm**2 / s,
                0.6042349293467888e-5 * cm**2 / s,
                1.2284727588198259e-5 * cm**2 / s,
                2.0323959442351844e-5 * cm**2 / s,
                4.064791888470369e-5 * cm**2 / s,
                1.3549306294901227e-5 * cm**2 / s,
                1.0161979721175922e-5 * cm**2 / s,
            ],
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        # Construct diffus_comp and all prerequisites
        state.diffus_comp

        assert_units_consistent(state.diffus_comp_constraint)

        param_sweeper = ParamSweeper(n_scen, state_values, output_values=target_values)
        with param_sweeper:
            for inputs, target in param_sweeper:
                solve_strongly_connected_components(state)

                # Make sure property equations have been converged
                assert number_large_residuals(state, tol=1e-8) == 0

                # Sanity check that inputs are properly set
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                # Make sure properties have been calculated as expected
                for var, val in target.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

    def test_therm_cond(self):
        m = self._make_model()
        state = m.fs.state

        n_scen = 8
        state_values = {
            "flow_mol": [1.0 * pyunits.mol / pyunits.s] * n_scen,
            "temperature": [
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                300.0 * pyunits.K,
                600.0 * pyunits.K,
                900.0 * pyunits.K,
                1200.0 * pyunits.K,
            ],
            "pressure": [1.0 * pyunits.bar] * n_scen,
            "mole_frac_comp[O2]": [1.0, 0.0, 0.0, 0.0, 0.25, 0.25, 0.25, 0.25],
            "mole_frac_comp[N2]": [0.0, 1.0, 0.0, 0.0, 0.25, 0.25, 0.25, 0.25],
            "mole_frac_comp[H2O]": [0.0, 0.0, 1.0, 0.0, 0.25, 0.25, 0.25, 0.25],
            "mole_frac_comp[CO2]": [0.0, 0.0, 0.0, 1.0, 0.25, 0.25, 0.25, 0.25],
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )

        u = pyunits.kJ / pyunits.m / pyunits.K / pyunits.s
        target_values = {
            "therm_cond": [
                8.490673044994837e-05 * u,
                7.803113104821915e-05 * u,
                0.0001245121534187936 * u,
                7.844692969560201e-05 * u,
                2.1981943936613706e-05 * u,
                4.567583423706824e-05 * u,
                6.946515568649932e-05 * u,
                9.30078254960681e-05 * u,
            ],
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        # Construct therm_cond and all prerequisites
        state.therm_cond

        param_sweeper = ParamSweeper(n_scen, state_values, output_values=target_values)
        with param_sweeper:
            for inputs, target in param_sweeper:
                solve_strongly_connected_components(state)

                # Make sure property equations have been converged
                assert number_large_residuals(state, tol=1e-8) == 0

                # Sanity check that inputs are properly set
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                # Make sure properties have been calculated as expected
                for var, val in target.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

    def test_cp(self):
        m = self._make_model()
        state = m.fs.state

        n_scen = 8
        state_values = {
            "flow_mol": [1.0 * pyunits.mol / pyunits.s] * n_scen,
            "temperature": [
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                300.0 * pyunits.K,
                600.0 * pyunits.K,
                900.0 * pyunits.K,
                1200.0 * pyunits.K,
            ],
            "pressure": [1.0 * pyunits.bar] * n_scen,
            "mole_frac_comp[O2]": [1.0, 0.0, 0.0, 0.0, 0.25, 0.25, 0.25, 0.25],
            "mole_frac_comp[N2]": [0.0, 1.0, 0.0, 0.0, 0.25, 0.25, 0.25, 0.25],
            "mole_frac_comp[H2O]": [0.0, 0.0, 1.0, 0.0, 0.25, 0.25, 0.25, 0.25],
            "mole_frac_comp[CO2]": [0.0, 0.0, 0.0, 1.0, 0.25, 0.25, 0.25, 0.25],
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )

        u = pyunits.kJ / pyunits.mol / pyunits.K
        kJkgK = pyunits.kJ / pyunits.kg / pyunits.K
        target_values = {
            "cp_mol_comp[O2]": [
                0.03566421043844448 * u,
                0.03566421043844444 * u,
                0.03566421043844444 * u,
                0.03566421043844444 * u,
                0.02408660519211111 * u,
                0.031970683705777776 * u,
                0.03435676292601234 * u,
                0.03566421043844444 * u,
            ],
            "cp_mol_comp[N2]": [
                0.03372177593533332 * u,
                0.03372177593533333 * u,
                0.03372177593533333 * u,
                0.03372177593533333 * u,
                0.03059729435133333 * u,
                0.030104019077333333 * u,
                0.03208929344525926 * u,
                0.03372177593533333 * u,
            ],
            "cp_mol_comp[H2O]": [
                0.0437510227322222 * u,
                0.04375102273222223 * u,
                0.04375102273222223 * u,
                0.04375102273222223 * u,
                0.03359738794555556 * u,
                0.036317861208888885 * u,
                0.039997715202839505 * u,
                0.04375102273222223 * u,
            ],
            "cp_mol_comp[CO2]": [
                0.05634605443600005 * u,
                0.056346054436000007 * u,
                0.056346054436000007 * u,
                0.056346054436000007 * u,
                0.037217621149000006 * u,
                0.047317934392 * u,
                0.053001289534111116 * u,
                0.056346054436000007 * u,
            ],
            "cp_mol": [
                0.03566421043844448 * u,
                0.03372177593533333 * u,
                0.04375102273222223 * u,
                0.056346054436000007 * u,
                0.0313747271595 * u,
                0.036427624596 * u,
                0.03986126527705556 * u,
                0.25
                * (
                    # Component values at 1200 K have been computed.
                    0.03566421043844444 * u
                    + 0.03372177593533333 * u
                    + 0.04375102273222223 * u
                    + 0.056346054436000007 * u
                ),
            ],
            "cp_mass": [
                1.1145065762013922 * kJkgK,
                1.2043491405476134 * kJkgK,
                2.4306123740123473 * kJkgK,
                1.280592146272725 * kJkgK,
                1.0286795790000056 * kJkgK,
                1.194348347409837 * kJkgK,
                1.3069267303952685 * kJkgK,
                1.3892054388688537 * kJkgK,
            ],
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        # Construct cp_mass and all prerequisites.
        # This constructs cp_mol and cp_mol_comp as well.
        state.cp_mass

        param_sweeper = ParamSweeper(n_scen, state_values, output_values=target_values)
        with param_sweeper:
            for inputs, target in param_sweeper:
                solve_strongly_connected_components(state)

                # Make sure property equations have been converged
                assert number_large_residuals(state, tol=1e-8) == 0

                # Sanity check that inputs are properly set
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                # Make sure properties have been calculated as expected
                for var, val in target.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

    def test_enth(self):
        m = self._make_model()
        state = m.fs.state

        n_scen = 7
        state_values = {
            "flow_mol": [1.0 * pyunits.mol / pyunits.s] * n_scen,
            "temperature": [
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                1200.0 * pyunits.K,
                # We do not test enthalpy at 300 K. The Shomate equation
                # we use is not valid below 500 K.
                # 300.0*pyunits.K,
                600.0 * pyunits.K,
                900.0 * pyunits.K,
                1200.0 * pyunits.K,
            ],
            "pressure": [1.0 * pyunits.bar] * n_scen,
            "mole_frac_comp[O2]": [1.0, 0.0, 0.0, 0.0, 0.25, 0.25, 0.25],
            "mole_frac_comp[N2]": [0.0, 1.0, 0.0, 0.0, 0.25, 0.25, 0.25],
            "mole_frac_comp[H2O]": [0.0, 0.0, 1.0, 0.0, 0.25, 0.25, 0.25],
            "mole_frac_comp[CO2]": [0.0, 0.0, 0.0, 1.0, 0.25, 0.25, 0.25],
        }
        state_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in state_values.items()
        )

        u = pyunits.kJ / pyunits.mol
        target_values = {
            "enth_mol_comp[O2]": [
                29.760175857866656 * u,
                29.760175857866656 * u,
                29.760175857866656 * u,
                29.760175857866656 * u,
                # 0.5175085434916653*u,
                9.248259058533336 * u,
                19.241674269713887 * u,
                29.760175857866656 * u,
            ],
            "enth_mol_comp[N2]": [
                28.1081423656 * u,
                28.1081423656 * u,
                28.1081423656 * u,
                28.1081423656 * u,
                # -0.021818752399997976*u,
                8.8939164816 * u,
                18.223311732266666 * u,
                28.1081423656 * u,
            ],
            "enth_mol_comp[H2O]": [
                34.505905041333335 * u,
                34.505905041333335 * u,
                34.505905041333335 * u,
                34.505905041333335 * u,
                # 0.06267505633334736*u,
                10.500564354666665 * u,
                21.939189237444452 * u,
                34.505905041333335 * u,
            ],
            "enth_mol_comp[CO2]": [
                44.474410900800024 * u,
                44.474410900800024 * u,
                44.474410900800024 * u,
                44.474410900800024 * u,
                # 0.06585135367498651*u,
                12.906441898799983 * u,
                28.031785067675003 * u,
                44.474410900800024 * u,
            ],
            "enth_mol": [
                29.760175857866656 * u,
                28.1081423656 * u,
                34.505905041333335 * u,
                44.474410900800024 * u,
                # 0.15605405027500296*u,
                10.387295448399996 * u,
                21.858990076775 * u,
                0.25
                * (
                    29.760175857866656 * u
                    + 28.1081423656 * u
                    + 34.505905041333335 * u
                    + 44.474410900800024 * u
                ),
            ],
        }
        target_values = ComponentMap(
            (state.find_component(name), values)
            for name, values in target_values.items()
        )

        # Construct enth_mol and all prerequisites, including enth_mol_comp
        state.enth_mol

        param_sweeper = ParamSweeper(n_scen, state_values, output_values=target_values)
        with param_sweeper:
            for inputs, target in param_sweeper:
                solve_strongly_connected_components(state)

                # Make sure property equations have been converged
                assert number_large_residuals(state, tol=1e-8) == 0

                # Sanity check that inputs are properly set
                for var, val in inputs.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)

                # Make sure properties have been calculated as expected
                for var, val in target.items():
                    val = value(pyunits.convert(val, var.get_units()))
                    assert var.value == pytest.approx(value(val), rel=1e-3)


if __name__ == "__main__":
    unittest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
General pure component electrolyte methods
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

from pyomo.environ import Var, units as pyunits

from idaes.core.util.misc import set_param_from_config


# -----------------------------------------------------------------------------
# Method for constant relative permittivity
# Relative permittivity was referred to as the dielectric constant in the past
class relative_permittivity_constant:
    """Methods for constant relative permittivity."""

    @staticmethod
    def build_parameters(cobj):
        cobj.relative_permittivity_liq_comp = Var(
            doc="Relative permittivity", units=pyunits.dimensionless
        )
        set_param_from_config(cobj, param="relative_permittivity_liq_comp")

    @staticmethod
    def return_expression(b, cobj, T):
        return cobj.relative_permittivity_liq_comp
